from typing import Any, Dict  # NOQA
from importlib import import_module
from logging import getLogger

from kedro.pipeline import Pipeline  # NOQA

from pipelinex import HatchDict
from .hooks_in_parameters_context import HooksInParametersContext


log = getLogger(__name__)


class FlexibleParametersContext(HooksInParametersContext):
    _params = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        params = self.params
        import_modules(params.pop("IMPORT", []))

        params = HatchDict(params).get()
        self._register_kedro_hooks(params.pop("HOOKS", None) or [])
        self._kedro_pipelines = params.pop("PIPELINES", None)
        self._kedro_run_config = params.pop("RUN_CONFIG", None) or {}
        self._params = params

    @property
    def params(self) -> Dict[str, Any]:
        if self._params is None:
            return super().params
        return self._params

    def _get_pipelines(self) -> Dict[str, Pipeline]:
        assert (
            self._kedro_pipelines
        ), "'PIPELINES' key or value is missing in parameters."
        return self._kedro_pipelines

    def run(self, *args, **kwargs):
        run_dict = self._kedro_run_config
        run_dict.update(kwargs)
        log.info("Run pipeline ({})".format(run_dict))
        return super().run(*args, **run_dict)


def import_modules(modules=None):
    if modules:
        if not isinstance(modules, list):
            modules = [modules]
        for module in modules:
            assert isinstance(module, str), "'{}' is not string.".format(module)
            import_module(module)



from importlib.util import find_spec
from logging import getLogger
from datetime import datetime, timedelta
import time
from typing import Any, Dict  # NOQA

from kedro.io import DataCatalog
from kedro.pipeline import Pipeline

from .mlflow_utils import (
    hook_impl,
    mlflow_start_run,
    mlflow_log_metrics,
    mlflow_log_params,
    mlflow_end_run,
)


log = getLogger(__name__)


def get_timestamp(dt=None, offset_hours=0, fmt="%Y-%m-%dT%H:%M:%S"):
    dt = dt or datetime.now()
    return (dt + timedelta(hours=offset_hours)).strftime(fmt)


def get_timestamp_int(dt=None, offset_hours=0):
    return int(get_timestamp(dt=dt, offset_hours=offset_hours, fmt="%Y%m%d%H%M%S"))


def get_timestamps(dt=None, offset_hours=0):
    dt = dt or datetime.now()
    timestamp = get_timestamp(dt, offset_hours=offset_hours)
    timestamp_int = get_timestamp_int(dt, offset_hours=offset_hours)
    return timestamp, timestamp_int


class MLflowBasicLoggerHook:
    """ Configures and logs duration time for the pipeline to MLflow
    """

    def __init__(
        self,
        enable_mlflow: bool = True,
        uri: str = None,
        experiment_name: str = None,
        artifact_location: str = None,
        run_name: str = None,
        offset_hours: float = None,
    ):
        """
        Args:
            enable_mlflow: Enable configuring and logging to MLflow.
            uri: The MLflow tracking server URI. 
                `uri` arg fed to:
                https://www.mlflow.org/docs/latest/python_api/mlflow.html#mlflow.set_tracking_uri
            experiment_name: The experiment name.
                `name` arg fed to:
                https://www.mlflow.org/docs/latest/python_api/mlflow.html#mlflow.create_experiment
            artifact_location: `artifact_location` arg fed to:
                https://www.mlflow.org/docs/latest/python_api/mlflow.html#mlflow.create_experiment
            run_name: Shown as 'Run Name' in MLflow UI. If `None`, Kedro's `run_id` is used.
            offset_hours: The offset hour (e.g. 0 for UTC+00:00) to log in MLflow. 

        """
        self.enable_mlflow = find_spec("mlflow") and enable_mlflow
        self.uri = uri
        self.experiment_name = experiment_name
        self.artifact_location = artifact_location
        self.run_name = run_name
        self.offset_hours = offset_hours or 0

    @hook_impl
    def before_pipeline_run(
        self, run_params: Dict[str, Any], pipeline: Pipeline, catalog: DataCatalog
    ):
        mlflow_start_run(
            uri=self.uri,
            experiment_name=self.experiment_name,
            artifact_location=self.artifact_location,
            run_name=self.run_name or run_params.get("run_id", ""),
            enable_mlflow=self.enable_mlflow,
        )

        run_params_renamed = {("___" + k): v for (k, v) in run_params.items()}
        mlflow_log_params(run_params_renamed, enable_mlflow=self.enable_mlflow)

        timestamp, timestamp_int = get_timestamps(offset_hours=self.offset_hours)
        time_dict = {"__time_begin": timestamp}
        time_int_dict = {"__time_begin": timestamp_int}

        mlflow_log_params(time_dict, enable_mlflow=self.enable_mlflow)
        mlflow_log_metrics(time_int_dict, enable_mlflow=self.enable_mlflow)

        self._time_begin = time.time()

    @hook_impl
    def after_pipeline_run(
        self, run_params: Dict[str, Any], pipeline: Pipeline, catalog: DataCatalog
    ):
        timestamp, timestamp_int = get_timestamps(offset_hours=self.offset_hours)
        time_dict = {"__time_end": timestamp}
        time_int_dict = {"__time_end": timestamp_int}

        mlflow_log_params(time_dict, enable_mlflow=self.enable_mlflow)
        mlflow_log_metrics(time_int_dict, enable_mlflow=self.enable_mlflow)

        self._time_end = time.time()
        self._time = self._time_end - self._time_begin

        time_dict = {"__time": self._time}
        mlflow_log_metrics(time_dict, enable_mlflow=self.enable_mlflow)

        mlflow_end_run(self.enable_mlflow)



from typing import Any, Dict  # NOQA
from logging import getLogger

from .context import KedroContext

log = getLogger(__name__)


class HooksInParametersContext(KedroContext):
    _hooks_in_params_read = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not hasattr(self, "_hook_manager"):

            class Hook:
                def after_catalog_created(self, *args, **kwargs):
                    pass

                def before_node_run(self, *args, **kwargs):
                    pass

                def after_node_run(self, *args, **kwargs):
                    pass

                def on_node_error(self, *args, **kwargs):
                    pass

                def before_pipeline_run(self, *args, **kwargs):
                    pass

                def after_pipeline_run(self, *args, **kwargs):
                    pass

                def on_pipeline_error(self, *args, **kwargs):
                    pass

            hook = Hook()

            class HookManager:
                def __init__(self):
                    self.hook = hook

                def is_registered(self, *args, **kwargs):
                    return False

                def register(self, *args, **kwargs):
                    pass

            self._hook_manager = HookManager()

    def _register_kedro_hooks(self, hooks):

        if not self._hooks_in_params_read:
            self._hooks_in_params_read = True

            if not isinstance(hooks, (list, tuple)):
                hooks = [hooks]

            for hook in hooks:
                if not self._hook_manager.is_registered(hook):
                    self._hook_manager.register(hook)



from pipelinex import log_time
from time import sleep
import logging

logging.basicConfig(level=logging.DEBUG)


@log_time
def sleeping_identity(inp):
    sleep(0.1)
    return inp


sleeping_identity(1)



from functools import wraps
from typing import Callable
import logging

log = logging.getLogger(__name__)


def dict_of_list_to_list_of_dict(dict_of_list):
    return [
        dict(zip(dict_of_list.keys(), vals)) for vals in zip(*dict_of_list.values())
    ]


def list_of_dict_to_dict_of_list(list_of_dict):
    return {k: [d[k] for d in list_of_dict] for k in list_of_dict[0]}


def dict_io(func: Callable) -> Callable:
    @wraps(func)
    def _dict_io(*args):
        keys = args[0].keys()

        out_dict = {}
        for key in keys:
            a = [e.get(key) for e in args]
            out = func(*a)
            out_dict[key] = out
            log.debug("{}: {}".format(key, out))

        if isinstance(out_dict[key], tuple):
            return tuple(dict_of_list_to_list_of_dict(out_dict))

        else:
            return out_dict

    return _dict_io


class DictToDict:
    module = None
    fn = None

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        assert isinstance(self.fn, str)

    def __call__(self, *args):
        def _apply_func_to_dicts(*args):

            module = self.module
            fn = self.fn
            kwargs = self.kwargs

            if module is None:
                fn = eval(fn)
            else:
                if isinstance(module, str):
                    module = eval(module)
                fn = getattr(module, fn)
            out = fn(*args, **kwargs)
            return out

        @dict_io
        def apply_func_to_dicts(*args):
            return _apply_func_to_dicts(*args)

        return apply_func_to_dicts(*args)


class ItemGetter:
    def __init__(self, item):
        self.item = item

    def __call__(self, d):
        return d[self.item]


class TransformCompose:
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, d):
        for t in self.transforms:
            d = t(d)
        return d



import copy
from typing import Any, Dict, Union
import numpy as np
from .csv_local import CSVLocalDataSet
import logging

log = logging.getLogger(__name__)


class EfficientCSVLocalDataSet(CSVLocalDataSet):
    """
    """

    DEFAULT_LOAD_ARGS = dict(
        engine="c", keep_default_na=False, na_values=[""], skiprows=0
    )  # type: Dict[str, Any]
    DEFAULT_PREVIEW_ARGS = dict(nrows=None, low_memory=False)  # type: Dict[str, Any]

    def __init__(
        self,
        *args,
        preview_args: Dict[str, Any] = None,
        margin: float = 100.0,
        verbose: Union[bool, int] = True,
        **kwargs
    ) -> None:
        """Creates a new instance of ``PandasDescribeDataSet`` pointing to a concrete
        filepath.

        Args:
            args: Positional arguments for ``CSVLocalDataSet``
            preview_args: Arguments passed on to ``df.describe``.
                See https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.describe.html for details.
            kwargs: Keyword arguments for ``CSVLocalDataSet``

        """
        super().__init__(*args, **kwargs)

        self._preview_args = copy.deepcopy(self.DEFAULT_PREVIEW_ARGS)
        if preview_args is not None:
            self._preview_args.update(preview_args)
        self._margin = margin
        self._verbose = verbose

    def _describe(self) -> Dict[str, Any]:
        return dict(
            filepath=self._filepath,
            preview_args=self._preview_args,
            load_args=self._load_args,
            save_args=self._save_args,
            version=self._version,
        )

    def _load(self) -> Any:

        load_args = self._load_args.copy()
        float_dtype = self._load_args.pop("float_dtype", "float16")
        assert float_dtype in {"float16", "float32", "float64"}
        int_dtype = self._load_args.pop("int_dtype", "int8")
        assert int_dtype in {"int8", "int16", "int32", "int64"}

        nrows = self._preview_args.get("nrows")

        self._load_args.update(self._preview_args)
        df = super()._load()

        dtypes_dict = _get_necessary_dtypes(
            df,
            margin=self._margin,
            float_dtype=float_dtype,
            int_dtype=int_dtype,
            verbose=self._verbose,
        )
        if nrows:
            load_args["dtype"] = dtypes_dict
            self._load_args = load_args
            df = super()._load()
        else:
            df = df.astype(dtype=dtypes_dict)
        return df


def dict_val_replace_except(
    d,  # type: dict
    to_except,  # type: Any
    new_value,  # type: Any
):
    return {k: (new_value if v != to_except else v) for k, v in d.items()}


def dict_string_val_prefix(
    d,  # type: dict
    prefix,  # type: Any
):
    return {k: (prefix + v) for k, v in d.items()}


def _get_necessary_dtypes(
    df, margin=100, float_dtype="float16", int_dtype="int8", verbose=True
):
    dtypes_dict = df.dtypes.to_dict()
    # dtypes_dict = {col: dtype.name for col, dtype in dtypes_dict.items()}

    for col, dtype in dtypes_dict.items():
        if dtype == np.object:
            dtypes_dict[col] = "object"  # np.object
        if dtype == np.float64:
            if float_dtype == "float16" and (
                (df[col].max() + margin) < np.finfo(np.float16).max
                and (df[col].min() - margin) > np.finfo(np.float16).min
            ):
                dtypes_dict[col] = "float16"  # np.float16
            elif float_dtype in {"float32", "float16"} and (
                (df[col].max() + margin) < np.finfo(np.float32).max
                and (df[col].min() - margin) > np.finfo(np.float32).min
            ):
                dtypes_dict[col] = "float32"  # np.float32
            else:
                dtypes_dict[col] = "float64"  # np.float64

        if dtype == np.int64:
            if int_dtype == "int8" and (
                (df[col].max() + margin) < np.iinfo(np.int8).max
                and (df[col].min() - margin) > np.iinfo(np.int8).min
            ):
                dtypes_dict[col] = "int8"  # np.int8
            if int_dtype == "int16" and (
                (df[col].max() + margin) < np.iinfo(np.int16).max
                and (df[col].min() - margin) > np.iinfo(np.int16).min
            ):
                dtypes_dict[col] = "float16"  # np.float16
            elif int_dtype in {"int32", "int16"} and (
                (df[col].max() + margin) < np.iinfo(np.int32).max
                and (df[col].min() - margin) > np.iinfo(np.int32).min
            ):
                dtypes_dict[col] = "int32"  # np.int32
            else:
                dtypes_dict[col] = "int64"  # np.int64

    if verbose:

        for dtype in [
            "float16",
            "float32",
            "float64",
            "int8",
            "int16",
            "int32",
            "int64",
            "object",
        ]:
            log.info(
                (
                    "{} is the minimum dtype for columns: \n{}".format(
                        dtype, [col for col, t in dtypes_dict.items() if t == dtype]
                    )
                )
            )

    return dtypes_dict



import torch
from ignite.utils import to_onehot


class ClassificationOutputTransform:
    def __init__(self, num_classes=None):
        self._num_classes = num_classes

    def __call__(self, output):
        if isinstance(output, tuple):
            y_pred, y = output
        elif isinstance(output, dict):
            y_pred = output["y_pred"]
            y = output["y"]
        else:
            raise ValueError

        if self._num_classes:
            y_pred = y_pred.clamp(min=0, max=self._num_classes - 1).long()
            y = y.clamp(min=0, max=self._num_classes - 1).long()
            y_pred = to_onehot(y_pred, self._num_classes)
        else:
            y_pred = y_pred.long()
            y = y.long()
        return y_pred, y



# Copyright 2020 QuantumBlack Visual Analytics Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND
# NONINFRINGEMENT. IN NO EVENT WILL THE LICENSOR OR OTHER CONTRIBUTORS
# BE LIABLE FOR ANY CLAIM, DAMAGES, OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF, OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# The QuantumBlack Visual Analytics Limited ("QuantumBlack") name and logo
# (either separately or in combination, "QuantumBlack Trademarks") are
# trademarks of QuantumBlack. The License does not grant you any right or
# license to the QuantumBlack Trademarks. You may not use the QuantumBlack
# Trademarks or any confusingly similar mark as a trademark for your product,
#     or use the QuantumBlack Trademarks in any other manner that might cause
# confusion in the marketplace, including but not limited to in advertising,
# on websites, or on software.
#
# See the License for the specific language governing permissions and
# limitations under the License.

"""``CSVLocalDataSet`` loads and saves data to a local csv file. The
underlying functionality is supported by pandas, so it supports all
allowed pandas options for loading and saving csv files.
"""
import copy
from pathlib import Path
from typing import Any, Dict

import pandas as pd

from ..core import (
    AbstractVersionedDataSet,
    Version,
)


class CSVLocalDataSet(AbstractVersionedDataSet):
    """``CSVLocalDataSet`` loads and saves data to a local csv file. The
    underlying functionality is supported by pandas, so it supports all
    allowed pandas options for loading and saving csv files.

    Example:
    ::

        >>> from kedro.io import CSVLocalDataSet
        >>> import pandas as pd
        >>>
        >>> data = pd.DataFrame({'col1': [1, 2], 'col2': [4, 5],
        >>>                      'col3': [5, 6]})
        >>> data_set = CSVLocalDataSet(filepath="test.csv",
        >>>                                  load_args=None,
        >>>                                  save_args={"index": False})
        >>> data_set.save(data)
        >>> reloaded = data_set.load()
        >>>
        >>> assert data.equals(reloaded)

    """

    DEFAULT_LOAD_ARGS = {}  # type: Dict[str, Any]
    DEFAULT_SAVE_ARGS = {"index": False}  # type: Dict[str, Any]

    def __init__(
        self,
        filepath: str,
        load_args: Dict[str, Any] = None,
        save_args: Dict[str, Any] = None,
        version: Version = None,
    ) -> None:
        """Creates a new instance of ``CSVLocalDataSet`` pointing to a concrete
        filepath.

        Args:
            filepath: path to a csv file.
            load_args: Pandas options for loading csv files.
                Here you can find all available arguments:
                https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_csv.html
                All defaults are preserved.
            save_args: Pandas options for saving csv files.
                Here you can find all available arguments:
                https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_csv.html
                All defaults are preserved, but "index", which is set to False.
            version: If specified, should be an instance of
                ``kedro.io.core.Version``. If its ``load`` attribute is
                None, the latest version will be loaded. If its ``save``
                attribute is None, save version will be autogenerated.

        Raises:
            ValueError: If 'filepath' looks like a remote path.
        """
        super().__init__(Path(filepath), version)

        # Handle default load and save arguments
        self._load_args = copy.deepcopy(self.DEFAULT_LOAD_ARGS)
        if load_args is not None:
            self._load_args.update(load_args)

        self._save_args = copy.deepcopy(self.DEFAULT_SAVE_ARGS)
        if save_args is not None:
            self._save_args.update(save_args)

    def _load(self) -> pd.DataFrame:
        load_path = Path(self._get_load_path())
        return pd.read_csv(load_path, **self._load_args)

    def _save(self, data: pd.DataFrame) -> None:
        save_path = Path(self._get_save_path())
        save_path.parent.mkdir(parents=True, exist_ok=True)
        data.to_csv(str(save_path), **self._save_args)

    def _exists(self) -> bool:
        path = self._get_load_path()
        return Path(path).is_file()

    def _describe(self) -> Dict[str, Any]:
        return dict(
            filepath=self._filepath,
            load_args=self._load_args,
            save_args=self._save_args,
            version=self._version,
        )



import copy
import logging
from pathlib import Path
from typing import Any, Dict  # NOQA

from .context import KedroContext
from kedro.io import DataCatalog
from pipelinex import HatchDict

log = logging.getLogger(__name__)


class FlexibleCatalogContext(KedroContext):
    """ Convert Kedrex's Syntactic Sugar to pure Kedro Catalog. """

    def _create_catalog(  # pylint: disable=no-self-use,too-many-arguments
        self,
        conf_catalog: Dict[str, Any],
        conf_creds: Dict[str, Any],
        save_version: str = None,
        journal: Any = None,
        load_versions: Dict[str, str] = None,
    ) -> DataCatalog:
        """DataCatalog instantiation.
        Allow whether to apply CachedDataSet using `cached` key.
        Returns:
            DataCatalog defined in `catalog.yml`.
        """
        conf_catalog = self._format_kedro_catalog(conf_catalog)
        return DataCatalog.from_config(
            conf_catalog,
            conf_creds,
            save_version=save_version,
            load_versions=load_versions,
        )

    def _format_kedro_catalog(self, conf_catalog):

        conf_catalog = HatchDict(conf_catalog).get()

        default_dict = {}
        if "/" in conf_catalog:
            default_dict = conf_catalog.pop("/")

        if "PIPELINE_JSON_TEXT" in conf_catalog:
            pipeline_json_text_dataset = conf_catalog.pop("PIPELINE_JSON_TEXT")
            assert isinstance(pipeline_json_text_dataset, dict)
            pipeline_json_text_dataset.setdefault(
                "type", "kedro.extras.datasets.text.TextDataSet"
            )
            self._pipeline_json_text_dataset = HatchDict(
                pipeline_json_text_dataset, obj_key="type"
            ).get()

        conf_catalog_processed = {}

        for ds_name, ds_dict_ in conf_catalog.items():
            ds_dict = copy.deepcopy(default_dict)
            if isinstance(ds_dict_, dict):
                ds_dict.update(ds_dict_)
            _check_type(ds_dict)
            ds_name, ds_dict = self._format_kedro_dataset(ds_name, ds_dict)
            conf_catalog_processed[ds_name] = ds_dict
        return conf_catalog_processed

    def _format_kedro_dataset(self, ds_name, ds_dict):
        ds_name, ds_dict = self._set_filepath(ds_name, ds_dict)
        ds_name, ds_dict = self._enable_caching(ds_name, ds_dict)
        return ds_name, ds_dict

    def _set_filepath(self, ds_name, ds_dict):
        if not any([(key in ds_dict) for key in ["filepath", "path", "url", "urls"]]):
            ds_dict["filepath"] = ds_name
            ds_name = Path(ds_name).stem
        return ds_name, ds_dict

    def _enable_caching(self, ds_name, ds_dict):
        cached = False
        if "cached" in ds_dict:
            cached = ds_dict.pop("cached")
        if cached and (ds_dict.get("type") != "kedro.io.CachedDataSet"):
            ds_dict = {
                "type": "kedro.io.CachedDataSet",
                "dataset": ds_dict,
            }
        return ds_name, ds_dict


def _check_type(ds_dict):
    type = ds_dict.get("type")
    if type and (not type.endswith("DataSet")):
        log.warning("type: '{}' does not end with 'DataSet'.".format(type))



import tempfile
import numpy as np
from pipelinex import ImagesLocalDataSet


def test_save_and_load():

    with tempfile.TemporaryDirectory() as dir:

        ds = ImagesLocalDataSet(
            path=dir + "/foobar_images", save_args={"suffix": ".jpg"}
        )

        image_dict = {
            "foo_image": 255 * np.ones((480, 640, 3), dtype=np.uint8),
            "bar_image": np.zeros((100, 200, 3), dtype=np.uint8),
        }

        ds._save(image_dict)
        loaded_image_dict = ds._load()

        for name in image_dict.keys():
            assert (image_dict[name] == loaded_image_dict[name]).all()


if __name__ == "__main__":
    test_save_and_load()



import logging
import time
from functools import wraps
from typing import Callable

log = logging.getLogger(__name__)


def _func_full_name(func: Callable):
    return getattr(func, "__qualname__", repr(func))


def _human_readable_time(elapsed: float):  # pragma: no cover
    mins, secs = divmod(elapsed, 60)
    hours, mins = divmod(mins, 60)

    if hours > 0:
        message = "%dh%02dm%02ds" % (hours, mins, secs)
    elif mins > 0:
        message = "%dm%02ds" % (mins, secs)
    elif secs >= 1:
        message = "%.2fs" % secs
    else:
        message = "%.0fms" % (secs * 1000.0)

    return message


def log_time(func: Callable) -> Callable:
    """A function decorator which logs the time taken for executing a function.

    Args:
        func: The function to be logged.

    Returns:
        A wrapped function, which will execute the provided function and log
        the running time.

    """

    @wraps(func)
    def with_time(*args, **kwargs):
        log = logging.getLogger(__name__)
        t_start = time.time()
        result = func(*args, **kwargs)
        t_end = time.time()
        elapsed = t_end - t_start

        log.info(
            "Running %r took %s [%.3fs]",
            _func_full_name(func),
            _human_readable_time(elapsed),
            elapsed,
        )
        return result

    return with_time



from functools import wraps
import pandas as pd
from typing import Callable, List, Union

import logging

log = logging.getLogger(__name__)


def log_df_summary(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(df, *args, **kwargs):
        log = logging.getLogger(__name__)
        if isinstance(df, pd.DataFrame):
            shape_before = df.shape
            cols_before = df.columns.to_list()
            log.info(
                "{}".format(
                    dict(
                        DF_shape_before="{}".format(shape_before),
                        func="{}".format(getattr(func, "__qualname__", repr(func))),
                    )
                )
            )
        df = func(df, *args, **kwargs)
        if isinstance(df, pd.DataFrame):
            shape_after = df.shape
            cols_after = df.columns.to_list()
            cols_added = [col for col in cols_after if col not in cols_before]
            cols_removed = [col for col in cols_before if col not in cols_after]
            log.info(
                "{}".format(
                    dict(
                        DF_shape_after="{}".format(shape_after),
                        Columns_added="{}".format(cols_added),
                        Columns_removed="{}".format(cols_removed),
                    )
                )
            )
        return df

    return wrapper


def df_set_index(cols: Union[List[str], str],) -> Callable:
    """ decorator with arguments """
    if not isinstance(cols, list):
        cols = [cols]

    def decorator(func: Callable) -> Callable:
        """ decorator without arguments """

        @wraps(func)
        def wrapper(df, parameters, *args, **kwargs):
            if isinstance(df, pd.DataFrame):
                for col in cols:
                    if col not in df.columns:
                        log.warning("Could not find column: ".format(col))
                cols_ = [col for col in cols if col in df.columns]
                df.set_index(keys=cols_, inplace=True)
            df = func(df, parameters, *args, **kwargs)
            if isinstance(df, pd.DataFrame):
                df.reset_index(inplace=True)
            return df

        return wrapper

    return decorator


def total_seconds_to_datetime(
    cols: Union[List[str], str], origin: str = "1970-01-01"
) -> Callable:
    """ decorator with arguments """
    if not isinstance(cols, list):
        cols = [cols]

    def decorator(func: Callable) -> Callable:
        """ decorator without arguments """

        @wraps(func)
        def wrapper(df, parameters, *args, **kwargs):
            if isinstance(df, pd.DataFrame):
                for col in cols:
                    if col not in df.columns:
                        log.warning("Could not find column: ".format(col))
                cols_ = [col for col in cols if col in df.columns]
                for col in cols_:
                    df.loc[:, col] = pd.to_datetime(
                        df[col], unit="s", origin=pd.Timestamp(origin)
                    )
            df = func(df, parameters, *args, **kwargs)
            if isinstance(df, pd.DataFrame):
                for col in cols_:
                    df.loc[:, col] = (df[col] - pd.Timestamp(origin)).dt.total_seconds()
            return df

        return wrapper

    return decorator



import importlib
from typing import Any, Union, List, Iterable  # NOQA


class HatchDict:
    def __init__(
        self,
        egg,  # type: Union[dict, List]
        lookup={},  # type: dict
        support_nested_keys=True,  # type: bool
        self_lookup_key="$",  # type: str
        support_import=True,  # type: bool
        additional_import_modules=["pipelinex"],  # type: Union[List, str]
        obj_key="=",  # type: str
        eval_parentheses=True,  # type: bool
    ):
        # type: (...) -> None

        assert egg.__class__.__name__ in {"dict", "list"}
        assert lookup.__class__.__name__ in {"dict"}
        assert support_nested_keys.__class__.__name__ in {"bool"}
        assert self_lookup_key.__class__.__name__ in {"str"}
        assert additional_import_modules.__class__.__name__ in {"list", "str"}
        assert obj_key.__class__.__name__ in {"str"}

        aug_egg = {}
        if isinstance(egg, dict):
            if support_nested_keys:
                aug_egg = dot_flatten(egg)
            aug_egg.update(egg)
        self.aug_egg = aug_egg

        self.egg = egg

        self.lookup = {}

        self.lookup.update(_builtin_funcs())
        self.lookup.update(lookup)

        self.self_lookup_key = self_lookup_key
        self.support_import = support_import
        self.additional_import_modules = (
            [additional_import_modules]
            if isinstance(additional_import_modules, str)
            else additional_import_modules or [__name__]
        )
        self.obj_key = obj_key
        self.eval_parentheses = eval_parentheses

        self.warmed_egg = None
        self.snapshot = None

    def get(
        self,
        key=None,  # type: Union[str, int]
        default=None,  # type: Any
        lookup={},  # type: dict
    ):
        # type: (...) -> Any

        assert (key is None) or (
            key.__class__.__name__ in {"str", "int",}
        ), "Received key: {}".format(key)
        assert lookup.__class__.__name__ in {"dict"}, "Received lookup: s{}".format(
            lookup
        )

        if key is None:
            d = self.egg
        else:
            if isinstance(self.egg, dict):
                d = self.aug_egg.get(key, default)
            if isinstance(self.egg, list):
                assert isinstance(key, int)
                d = self.egg[key] if (0 <= key < len(self.egg)) else default

        if self.self_lookup_key:
            s = dict()
            while d != s:
                d, s = _dfs_apply(
                    d_input=d,
                    hatch_args=dict(lookup=self.aug_egg, obj_key=self.self_lookup_key),
                )
            self.warmed_egg = d

        if self.eval_parentheses:
            d, s = _dfs_apply(
                d_input=d, hatch_args=dict(eval_parentheses=self.eval_parentheses)
            )
            self.warmed_egg = d

        lookup_input = {}
        lookup_input.update(self.lookup)
        lookup_input.update(lookup)

        if isinstance(self.egg, dict):
            forcing_module = self.egg.get("FORCING_MODULE", "")
            module_aliases = self.egg.get("MODULE_ALIASES", {})

        for m in self.additional_import_modules:
            d, s = _dfs_apply(
                d_input=d,
                hatch_args=dict(
                    lookup=lookup_input,
                    support_import=self.support_import,
                    default_module=m,
                    forcing_module=forcing_module,
                    module_aliases=module_aliases,
                    obj_key=self.obj_key,
                ),
            )
        self.snapshot = s
        return d

    def get_params(self):
        return self.snapshot

    def keys(self):
        return self.egg.keys()

    def items(self):
        assert isinstance(self.egg, dict)
        return [(k, self.get(k)) for k in self.egg.keys()]


def _dfs_apply(
    d_input,  # type: Any
    hatch_args,  # type: dict
):
    # type: (...) -> Any

    eval_parentheses = hatch_args.get("eval_parentheses", False)  # type: bool
    lookup = hatch_args.get("lookup", dict())  # type: dict
    support_import = hatch_args.get("support_import", False)  # type: bool
    default_module = hatch_args.get("default_module", "")  # type: str
    forcing_module = hatch_args.get("forcing_module", "")  # type: str
    module_aliases = hatch_args.get("module_aliases", {})  # type: dict
    obj_key = hatch_args.get("obj_key", "=")  # type: str

    d = d_input
    s = d_input

    if isinstance(d_input, dict):

        obj_str = d_input.get(obj_key)

        d, s = {}, {}
        for k, v in d_input.items():
            d[k], s[k] = _dfs_apply(v, hatch_args)

        if obj_str:
            if obj_str in lookup:
                a = lookup.get(obj_str)
                d = _hatch(d, a, obj_key=obj_key)
            elif support_import:
                if forcing_module:
                    obj_path_list = obj_str.rsplit(".", 1)
                    obj_str = "{}.{}".format(forcing_module, obj_path_list[-1])
                if module_aliases:
                    obj_path_list = obj_str.rsplit(".", 1)
                    if len(obj_path_list) == 2 and obj_path_list[0] in module_aliases:
                        module_alias = module_aliases.get(obj_path_list[0])
                        if module_alias is None:
                            obj_path_list.pop(0)
                        else:
                            obj_path_list[0] = module_alias
                        obj_str = ".".join(obj_path_list)
                a = load_obj(obj_str, default_obj_path=default_module)
                d = _hatch(d, a, obj_key=obj_key)

    if isinstance(d_input, list):

        d, s = [], []
        for v in d_input:
            _d, _s = _dfs_apply(v, hatch_args)
            d.append(_d)
            s.append(_s)

    if isinstance(d_input, str):
        if (
            eval_parentheses
            and len(d_input) >= 2
            and d_input[0] == "("
            and d_input[-1] == ")"
        ):
            d = eval(d)

    return d, s


def _hatch(
    d,  # type: dict
    a,  # type: Any
    obj_key="=",  # type: str
    pos_arg_key="_",  # type: str
    attr_key=".",  # type: str
):
    d.pop(obj_key)
    if d:
        assert callable(a), "{} is not callable.".format(a)

        pos_args = d.pop(pos_arg_key, None)
        if pos_args is None:
            pos_args = []
        if not isinstance(pos_args, list):
            pos_args = [pos_args]

        attribute_name = d.pop(attr_key, None)
        for k in d:
            assert isinstance(
                k, str
            ), "Non-string key '{}' in '{}' is not valid for callable: '{}'.".format(
                k, d, a.__name__
            )
        d = a(*pos_args, **d)
        if attribute_name:
            d = getattr(d, attribute_name)
            # if isinstance(d, MethodType):
            #     d = lambda *args: d(args[0])
    else:
        d = a
    return d


def dot_flatten(d):
    try:
        from flatten_dict import flatten
    except Exception:
        return d

    def dot_reducer(k1, k2):
        return k1 + "." + k2 if k1 else k2

    return flatten(d, reducer=dot_reducer)


def pass_(*argsignore, **kwargsignore):
    return None


def pass_through(*args, **kwargs):
    return args[0] if args else list(kwargs.values())[0] if kwargs else None


class ToPipeline:
    def __init__(self, *args):
        if len(args) == 1:
            args = args[0]
        self.args = args

    def __call__(self):
        return self.args


class Construct:
    def __init__(self, obj):
        self.obj = obj

    def __call__(self, *args, **kwargs):
        return self.obj(*args, **kwargs)


class Method:
    method = None

    def __init__(self, *args, **kwargs):
        if self.method is None:
            self.method = kwargs.pop("method")
        self.args = args
        self.kwargs = kwargs

    def __call__(self, d):
        if isinstance(d, dict):
            d = HatchDict(d)
        attr = getattr(d, self.method, None)
        if callable(attr):
            return attr(*self.args, **self.kwargs)
        else:
            return d


class Get(Method):
    method = "get"


def feed(func, args):
    assert callable(func)
    if isinstance(args, dict):
        posargs = args.pop("_", [])
        kwargs = args
    elif isinstance(args, (list, tuple)):
        posargs = args
        kwargs = dict()
    else:
        posargs = [args]
        kwargs = dict()

    def _feed(*argsignore, **kwargsignore):
        return func(*posargs, **kwargs)

    return _feed


def _builtin_funcs():
    return dict(
        pass_=pass_,
        pass_through=pass_through,
        ToPipeline=ToPipeline,
        Construct=Construct,
        Method=Method,
        Get=Get,
    )


"""
Copyright 2018-2019 QuantumBlack Visual Analytics Limited
regarding `load_obj` function copied from
https://github.com/quantumblacklabs/kedro/blob/0.15.4/kedro/utils.py
"""


def load_obj(obj_path: str, default_obj_path: str = "") -> Any:
    """Extract an object from a given path.
        Args:
            obj_path: Path to an object to be extracted, including the object name.
            default_obj_path: Default object path.
        Returns:
            Extracted object.
        Raises:
            AttributeError: When the object does not have the given named attribute.
    """
    obj_path_list = obj_path.rsplit(".", 1)
    obj_path = obj_path_list.pop(0) if len(obj_path_list) > 1 else default_obj_path
    obj_name = obj_path_list[0]
    module_obj = importlib.import_module(obj_path)
    if not hasattr(module_obj, obj_name):
        raise AttributeError(
            "Object `{}` cannot be loaded from `{}`.".format(obj_name, obj_path)
        )
    return getattr(module_obj, obj_name)



import kedro

from .flexible_catalog_context import FlexibleCatalogContext
from .flexible_parameters_context import FlexibleParametersContext
from .flexible_run_context import FlexibleRunContext


class FlexibleContext(
    FlexibleParametersContext, FlexibleCatalogContext, FlexibleRunContext,
):
    project_name = "KedroProject"
    project_version = kedro.__version__


class MLflowFlexibleContext(FlexibleContext):
    """ Deprecated alias for FlexibleContext for backward-compatibility """

    pass



import re
from codecs import open
from os import path

from setuptools import find_packages, setup

name = "pipelinex"
here = path.abspath(path.dirname(__file__))

with open(path.join(here, "src", name, "__init__.py"), encoding="utf-8") as f:
    result = re.search(r'__version__ = ["\']([^"\']+)', f.read())

    if not result:
        raise ValueError("Can't find the version")

    version = result.group(1)

setup(
    name=name,
    version=version,
    description="Tool to build production-ready pipelines for experimentation with Kedro and MLflow",
    license="Apache Software License (Apache 2.0)",
    long_description="Please see: https://github.com/Minyus/pipelinex",
    long_description_content_type="text/markdown",
    url="https://github.com/Minyus/pipelinex",
    packages=find_packages(where="src", exclude=["tests"]),
    package_dir={"": "src"},
    test_suite="tests",
    install_requires=[],
    author="Yusuke Minami",
    author_email="me@minyus.github.com",
    zip_safe=False,
    keywords="pipelines, machine learning, deep learning, data science, data engineering",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
)



# Copyright 2020 QuantumBlack Visual Analytics Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND
# NONINFRINGEMENT. IN NO EVENT WILL THE LICENSOR OR OTHER CONTRIBUTORS
# BE LIABLE FOR ANY CLAIM, DAMAGES, OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF, OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# The QuantumBlack Visual Analytics Limited ("QuantumBlack") name and logo
# (either separately or in combination, "QuantumBlack Trademarks") are
# trademarks of QuantumBlack. The License does not grant you any right or
# license to the QuantumBlack Trademarks. You may not use the QuantumBlack
# Trademarks or any confusingly similar mark as a trademark for your product,
#     or use the QuantumBlack Trademarks in any other manner that might cause
# confusion in the marketplace, including but not limited to in advertising,
# on websites, or on software.
#
# See the License for the specific language governing permissions and
# limitations under the License.


from importlib.util import find_spec

if find_spec("kedro"):
    from kedro.io.core import *
else:

    """This module provides a set of classes which underpin the data loading and
    saving functionality provided by ``kedro.io``.
    """

    import abc
    import copy
    import logging
    import re
    import warnings
    from collections import namedtuple
    from datetime import datetime, timezone
    from functools import lru_cache
    from glob import iglob
    from pathlib import Path, PurePath
    from typing import Any, Callable, Dict, List, Optional, Tuple, Type
    from urllib.parse import urlsplit

    from pipelinex.hatch_dict.hatch_dict import load_obj

    warnings.simplefilter("default", DeprecationWarning)

    VERSION_FORMAT = "%Y-%m-%dT%H.%M.%S.%fZ"
    VERSIONED_FLAG_KEY = "versioned"
    VERSION_KEY = "version"
    HTTP_PROTOCOLS = ("http", "https")
    PROTOCOL_DELIMITER = "://"
    CLOUD_PROTOCOLS = ("s3", "gcs", "gs", "adl", "abfs")

    class DataSetError(Exception):
        """``DataSetError`` raised by ``AbstractDataSet`` implementations
        in case of failure of input/output methods.

        ``AbstractDataSet`` implementations should provide instructive
        information in case of failure.
        """

        pass

    class DataSetNotFoundError(DataSetError):
        """``DataSetNotFoundError`` raised by ``DataCatalog`` class in case of
        trying to use a non-existing data set.
        """

        pass

    class DataSetAlreadyExistsError(DataSetError):
        """``DataSetAlreadyExistsError`` raised by ``DataCatalog`` class in case
        of trying to add a data set which already exists in the ``DataCatalog``.
        """

        pass

    class VersionNotFoundError(DataSetError):
        """``VersionNotFoundError`` raised by ``AbstractVersionedDataSet`` implementations
        in case of no load versions available for the data set.
        """

        pass

    class AbstractDataSet(abc.ABC):
        """``AbstractDataSet`` is the base class for all data set implementations.
        All data set implementations should extend this abstract class
        and implement the methods marked as abstract.

        Example:
        ::

            >>> from kedro.io import AbstractDataSet
            >>> import pandas as pd
            >>>
            >>> class MyOwnDataSet(AbstractDataSet):
            >>>     def __init__(self, param1, param2):
            >>>         self._param1 = param1
            >>>         self._param2 = param2
            >>>
            >>>     def _load(self) -> pd.DataFrame:
            >>>         print("Dummy load: {}".format(self._param1))
            >>>         return pd.DataFrame()
            >>>
            >>>     def _save(self, df: pd.DataFrame) -> None:
            >>>         print("Dummy save: {}".format(self._param2))
            >>>
            >>>     def _describe(self):
            >>>         return dict(param1=self._param1, param2=self._param2)
        """

        @classmethod
        def from_config(
            cls: Type,
            name: str,
            config: Dict[str, Any],
            load_version: str = None,
            save_version: str = None,
        ) -> "AbstractDataSet":
            """Create a data set instance using the configuration provided.

            Args:
                name: Data set name.
                config: Data set config dictionary.
                load_version: Version string to be used for ``load`` operation if
                    the data set is versioned. Has no effect on the data set
                    if versioning was not enabled.
                save_version: Version string to be used for ``save`` operation if
                    the data set is versioned. Has no effect on the data set
                    if versioning was not enabled.

            Returns:
                An instance of an ``AbstractDataSet`` subclass.

            Raises:
                DataSetError: When the function fails to create the data set
                    from its config.

            """
            try:
                class_obj, config = parse_dataset_definition(
                    config, load_version, save_version
                )
            except Exception as ex:
                raise DataSetError(
                    "An exception occurred when parsing config "
                    "for DataSet `{}`:\n{}".format(name, str(ex))
                )

            try:
                data_set = class_obj(**config)  # type: ignore
            except TypeError as err:
                raise DataSetError(
                    "\n{}.\nDataSet '{}' must only contain "
                    "arguments valid for the constructor "
                    "of `{}.{}`.".format(
                        str(err), name, class_obj.__module__, class_obj.__qualname__
                    )
                )
            except Exception as err:
                raise DataSetError(
                    "\n{}.\nFailed to instantiate DataSet "
                    "'{}' of type `{}.{}`.".format(
                        str(err), name, class_obj.__module__, class_obj.__qualname__
                    )
                )
            return data_set

        @property
        def _logger(self) -> logging.Logger:
            return logging.getLogger(__name__)

        def load(self) -> Any:
            """Loads data by delegation to the provided load method.

            Returns:
                Data returned by the provided load method.

            Raises:
                DataSetError: When underlying load method raises error.

            """

            self._logger.debug("Loading %s", str(self))

            try:
                return self._load()
            except DataSetError:
                raise
            except Exception as exc:
                # This exception handling is by design as the composed data sets
                # can throw any type of exception.
                message = "Failed while loading data from data set {}.\n{}".format(
                    str(self), str(exc)
                )
                raise DataSetError(message) from exc

        def save(self, data: Any) -> None:
            """Saves data by delegation to the provided save method.

            Args:
                data: the value to be saved by provided save method.

            Raises:
                DataSetError: when underlying save method raises error.

            """

            if data is None:
                raise DataSetError("Saving `None` to a `DataSet` is not allowed")

            try:
                self._logger.debug("Saving %s", str(self))
                self._save(data)
            except DataSetError:
                raise
            except Exception as exc:
                message = "Failed while saving data to data set {}.\n{}".format(
                    str(self), str(exc)
                )
                raise DataSetError(message) from exc

        def __str__(self):
            def _to_str(obj, is_root=False):
                """Returns a string representation where
                1. The root level (i.e. the DataSet.__init__ arguments) are
                formatted like DataSet(key=value).
                2. Dictionaries have the keys alphabetically sorted recursively.
                3. Empty dictionaries and None values are not shown.
                """

                fmt = "{}={}" if is_root else "'{}': {}"  # 1

                if isinstance(obj, dict):
                    sorted_dict = sorted(
                        obj.items(), key=lambda pair: str(pair[0])
                    )  # 2

                    text = ", ".join(
                        fmt.format(key, _to_str(value))  # 2
                        for key, value in sorted_dict
                        if value or isinstance(value, bool)
                    )  # 3

                    return text if is_root else "{" + text + "}"  # 1

                # not a dictionary
                return str(obj)

            return "{}({})".format(type(self).__name__, _to_str(self._describe(), True))

        @abc.abstractmethod
        def _load(self) -> Any:
            raise NotImplementedError(
                "`{}` is a subclass of AbstractDataSet and"
                "it must implement the `_load` method".format(self.__class__.__name__)
            )

        @abc.abstractmethod
        def _save(self, data: Any) -> None:
            raise NotImplementedError(
                "`{}` is a subclass of AbstractDataSet and"
                "it must implement the `_save` method".format(self.__class__.__name__)
            )

        @abc.abstractmethod
        def _describe(self) -> Dict[str, Any]:
            raise NotImplementedError(
                "`{}` is a subclass of AbstractDataSet and"
                "it must implement the `_describe` method".format(
                    self.__class__.__name__
                )
            )

        def exists(self) -> bool:
            """Checks whether a data set's output already exists by calling
            the provided _exists() method.

            Returns:
                Flag indicating whether the output already exists.

            Raises:
                DataSetError: when underlying exists method raises error.

            """
            try:
                self._logger.debug("Checking whether target of %s exists", str(self))
                return self._exists()
            except Exception as exc:
                message = "Failed during exists check for data set {}.\n{}".format(
                    str(self), str(exc)
                )
                raise DataSetError(message) from exc

        def _exists(self) -> bool:
            self._logger.warning(
                "`exists()` not implemented for `%s`. Assuming output does not exist.",
                self.__class__.__name__,
            )
            return False

        def release(self) -> None:
            """Release any cached data.

            Raises:
                DataSetError: when underlying release method raises error.

            """
            try:
                self._logger.debug("Releasing %s", str(self))
                self._release()
            except Exception as exc:
                message = "Failed during release for data set {}.\n{}".format(
                    str(self), str(exc)
                )
                raise DataSetError(message) from exc

        def _release(self) -> None:
            pass

        def _copy(self, **overwrite_params) -> "AbstractDataSet":
            dataset_copy = copy.deepcopy(self)
            for name, value in overwrite_params.items():
                setattr(dataset_copy, name, value)
            return dataset_copy

    def generate_timestamp() -> str:
        """Generate the timestamp to be used by versioning.

        Returns:
            String representation of the current timestamp.

        """
        current_ts = datetime.now(tz=timezone.utc).strftime(VERSION_FORMAT)
        return current_ts[:-4] + current_ts[-1:]  # Don't keep microseconds

    class Version(namedtuple("Version", ["load", "save"])):
        """This namedtuple is used to provide load and save versions for versioned
        data sets. If ``Version.load`` is None, then the latest available version
        is loaded. If ``Version.save`` is None, then save version is formatted as
        YYYY-MM-DDThh.mm.ss.sssZ of the current timestamp.
        """

        __slots__ = ()

    _CONSISTENCY_WARNING = (
        "Save version `{}` did not match load version `{}` for {}. This is strongly "
        "discouraged due to inconsistencies it may cause between `save` and "
        "`load` operations. Please refrain from setting exact load version for "
        "intermediate data sets where possible to avoid this warning."
    )

    _DEFAULT_PACKAGES = ["kedro.io.", "kedro.extras.datasets.", ""]

    def parse_dataset_definition(
        config: Dict[str, Any], load_version: str = None, save_version: str = None
    ) -> Tuple[Type[AbstractDataSet], Dict[str, Any]]:
        """Parse and instantiate a dataset class using the configuration provided.

        Args:
            config: Data set config dictionary. It *must* contain the `type` key
                with fully qualified class name.
            load_version: Version string to be used for ``load`` operation if
                    the data set is versioned. Has no effect on the data set
                    if versioning was not enabled.
            save_version: Version string to be used for ``save`` operation if
                the data set is versioned. Has no effect on the data set
                if versioning was not enabled.

        Raises:
            DataSetError: If the function fails to parse the configuration provided.

        Returns:
            2-tuple: (Dataset class object, configuration dictionary)
        """
        save_version = save_version or generate_timestamp()
        config = copy.deepcopy(config)

        if "type" not in config:
            raise DataSetError("`type` is missing from DataSet catalog configuration")

        class_obj = config.pop("type")
        if isinstance(class_obj, str):
            if len(class_obj.strip(".")) != len(class_obj):
                raise DataSetError(
                    "`type` class path does not support relative "
                    "paths or paths ending with a dot."
                )

            class_paths = (prefix + class_obj for prefix in _DEFAULT_PACKAGES)

            trials = (_load_obj(class_path) for class_path in class_paths)
            try:
                class_obj = next(obj for obj in trials if obj is not None)
            except StopIteration:
                raise DataSetError("Class `{}` not found.".format(class_obj))

        if not issubclass(class_obj, AbstractDataSet):
            raise DataSetError(
                "DataSet type `{}.{}` is invalid: all data set types must extend "
                "`AbstractDataSet`.".format(
                    class_obj.__module__, class_obj.__qualname__
                )
            )

        if VERSION_KEY in config:
            # remove "version" key so that it's not passed
            # to the "unversioned" data set constructor
            message = (
                "`%s` attribute removed from data set configuration since it is a "
                "reserved word and cannot be directly specified"
            )
            logging.getLogger(__name__).warning(message, VERSION_KEY)
            del config[VERSION_KEY]
        if config.pop(VERSIONED_FLAG_KEY, False):  # data set is versioned
            config[VERSION_KEY] = Version(load_version, save_version)

        return class_obj, config

    def _load_obj(class_path: str) -> Optional[object]:
        try:
            class_obj = load_obj(class_path)
        except ModuleNotFoundError as error:
            if error.name is None or error.name in class_path:
                return None
            # class_obj was successfully loaded, but some dependencies are missing.
            raise DataSetError(
                f"{error} for {class_path}. Please see the documentation on how to "
                f"install relevant dependencies for {class_path}:\n"
                f"https://kedro.readthedocs.io/en/stable/02_getting_started/"
                f"02_install.html#optional-dependencies"
            )
        except (AttributeError, ValueError):
            return None

        return class_obj

    def _local_exists(filepath: str) -> bool:  # SKIP_IF_NO_SPARK
        filepath = Path(filepath)
        return filepath.exists() or any(par.is_file() for par in filepath.parents)

    class AbstractVersionedDataSet(AbstractDataSet, abc.ABC):
        """
        ``AbstractVersionedDataSet`` is the base class for all versioned data set
        implementations. All data sets that implement versioning should extend this
        abstract class and implement the methods marked as abstract.

        Example:
        ::

            >>> from kedro.io import AbstractVersionedDataSet
            >>> import pandas as pd
            >>>
            >>>
            >>> class MyOwnDataSet(AbstractVersionedDataSet):
            >>>     def __init__(self, param1, param2, filepath, version):
            >>>         super().__init__(filepath, version)
            >>>         self._param1 = param1
            >>>         self._param2 = param2
            >>>
            >>>     def _load(self) -> pd.DataFrame:
            >>>         load_path = self._get_load_path()
            >>>         return pd.read_csv(load_path)
            >>>
            >>>     def _save(self, df: pd.DataFrame) -> None:
            >>>         save_path = self._get_save_path()
            >>>         df.to_csv(str(save_path))
            >>>
            >>>     def _exists(self) -> bool:
            >>>         path = self._get_load_path()
            >>>         return path.is_file()
            >>>
            >>>     def _describe(self):
            >>>         return dict(version=self._version, param1=self._param1, param2=self._param2)
        """

        # pylint: disable=abstract-method

        def __init__(
            self,
            filepath: PurePath,
            version: Optional[Version],
            exists_function: Callable[[str], bool] = None,
            glob_function: Callable[[str], List[str]] = None,
        ):
            """Creates a new instance of ``AbstractVersionedDataSet``.

            Args:
                filepath: Path to file.
                version: If specified, should be an instance of
                    ``kedro.io.core.Version``. If its ``load`` attribute is
                    None, the latest version will be loaded. If its ``save``
                    attribute is None, save version will be autogenerated.
                exists_function: Function that is used for determining whether
                    a path exists in a filesystem.
                glob_function: Function that is used for finding all paths
                    in a filesystem, which match a given pattern.
            """
            self._filepath = filepath
            self._version = version
            self._exists_function = exists_function or _local_exists
            self._glob_function = glob_function or iglob

        @lru_cache(maxsize=None)
        def resolve_load_version(self) -> Optional[str]:
            """Compute and cache the version the dataset should be loaded with."""
            if not self._version:
                return None
            if self._version.load:
                return self._version.load

            # When load version is unpinned, fetch the most recent existing
            # version from the given path
            pattern = str(self._get_versioned_path("*"))
            version_paths = sorted(self._glob_function(pattern), reverse=True)
            most_recent = next(
                (path for path in version_paths if self._exists_function(path)), None
            )

            if not most_recent:
                raise VersionNotFoundError(
                    "Did not find any versions for {}".format(str(self))
                )

            return PurePath(most_recent).parent.name

        def _get_load_path(self) -> PurePath:
            if not self._version:
                # When versioning is disabled, load from original filepath
                return self._filepath

            load_version = self.resolve_load_version()
            return self._get_versioned_path(load_version)  # type: ignore

        @lru_cache(maxsize=None)
        def resolve_save_version(self) -> Optional[str]:
            """Compute and cache the version the dataset should be saved with."""
            if not self._version:
                return None
            return self._version.save or generate_timestamp()

        def _get_save_path(self) -> PurePath:
            if not self._version:
                # When versioning is disabled, return original filepath
                return self._filepath

            save_version = self.resolve_save_version()
            versioned_path = self._get_versioned_path(save_version)  # type: ignore

            if self._exists_function(str(versioned_path)):
                raise DataSetError(
                    "Save path `{}` for {} must not exist if versioning "
                    "is enabled.".format(versioned_path, str(self))
                )

            return versioned_path

        def _get_versioned_path(self, version: str) -> PurePath:
            return self._filepath / version / self._filepath.name

        def load(self) -> Any:
            self.resolve_load_version()  # Make sure last load version is set
            return super().load()

        def save(self, data: Any) -> None:
            save_version = (
                self.resolve_save_version()
            )  # Make sure last save version is set
            super().save(data)

            load_version = self.resolve_load_version()
            if load_version != save_version:
                warnings.warn(
                    _CONSISTENCY_WARNING.format(save_version, load_version, str(self))
                )

        def exists(self) -> bool:
            """Checks whether a data set's output already exists by calling
            the provided _exists() method.

            Returns:
                Flag indicating whether the output already exists.

            Raises:
                DataSetError: when underlying exists method raises error.

            """
            self._logger.debug("Checking whether target of %s exists", str(self))
            try:
                return self._exists()
            except VersionNotFoundError:
                return False
            except Exception as exc:  # SKIP_IF_NO_SPARK
                message = "Failed during exists check for data set {}.\n{}".format(
                    str(self), str(exc)
                )
                raise DataSetError(message) from exc

        def _release(self) -> None:
            super()._release()
            self.resolve_load_version.cache_clear()
            self.resolve_save_version.cache_clear()

    def _parse_filepath(filepath: str) -> Dict[str, str]:
        """Split filepath on protocol and path. Based on `fsspec.utils.infer_storage_options`.

        Args:
            filepath: Either local absolute file path or URL (s3://bucket/file.csv)

        Returns:
            Parsed filepath.
        """
        if (
            re.match(r"^[a-zA-Z]:[\\/]", filepath)
            or re.match(r"^[a-zA-Z0-9]+://", filepath) is None
        ):
            return {"protocol": "file", "path": filepath}

        parsed_path = urlsplit(filepath)
        protocol = parsed_path.scheme or "file"

        if protocol in HTTP_PROTOCOLS:
            return {"protocol": protocol, "path": filepath}

        path = parsed_path.path
        if protocol == "file":
            windows_path = re.match(r"^/([a-zA-Z])[:|]([\\/].*)$", path)
            if windows_path:
                path = "{}:{}".format(*windows_path.groups())

        options = {"protocol": protocol, "path": path}

        if parsed_path.netloc:
            if protocol in CLOUD_PROTOCOLS:
                host_with_port = parsed_path.netloc.rsplit("@", 1)[-1]
                host = host_with_port.rsplit(":", 1)[0]
                options["path"] = host + options["path"]

        return options

    def get_protocol_and_path(
        filepath: str, version: Version = None
    ) -> Tuple[str, str]:
        """Parses filepath on protocol and path.

        Args:
            filepath: raw filepath e.g.: `gcs://bucket/test.json`.
            version: instance of ``kedro.io.core.Version`` or None.

        Returns:
            Protocol and path.

        Raises:
            DataSetError: when protocol is http(s) and version is not None.
            Note: HTTP(s) dataset doesn't support versioning.
        """
        options_dict = _parse_filepath(filepath)
        path = options_dict["path"]
        protocol = options_dict["protocol"]

        if protocol in HTTP_PROTOCOLS:
            if version:
                raise DataSetError(
                    "HTTP(s) DataSet doesn't support versioning. "
                    "Please remove version flag from the dataset configuration."
                )
            path = path.split(PROTOCOL_DELIMITER, 1)[-1]

        return protocol, path

    def get_filepath_str(path: PurePath, protocol: str) -> str:
        """Returns filepath. Returns full filepath (with protocol) if protocol is HTTP(s).

        Args:
            path: filepath without protocol.
            protocol: protocol.

        Returns:
            Filepath string.
        """
        path = str(path)
        if protocol in HTTP_PROTOCOLS:
            path = "".join((protocol, PROTOCOL_DELIMITER, path))
        return path

    def validate_on_forbidden_chars(**kwargs):
        """Validate that string values do not include white-spaces or ;"""
        for key, value in kwargs.items():
            if " " in value or ";" in value:
                raise DataSetError(
                    "Neither white-space nor semicolon are allowed in `{}`.".format(key)
                )
